unit DAG;

(******************************************************************************
 * Unit:          DAG
 *
 * Purpose:       Implement components for building visual representations of
 *                directed acylclic graphs (DAGs).
 *
 * Components:    TDAGNode       Visual node (colored circle)
 *                TDAGEdge       Visual edge (line segment joining two nodes)
 *                TDAGView       Scrollable view of a DAG
 *
 * Related
 *    types:      TDAGNodeList   Dynamic vector of TDAGNodes
 *                TDAGEdgeList   Dynamic vector of TDAGEdges
 *
 * Author:        Tres Seaver, Palladion Software
 *
 * History:       1 May 1995     Initial version
 ******************************************************************************)

{##############################################################################}
	interface
{##############################################################################}

uses Controls, Classes, Dialogs, Forms, Graphics, Menus, WinProcs, WinTypes;

const
	DAGNodeRadius = 10;
    DAGEdgeWidth = 1;

type

TDAGNode = class;
TDAGEdge = class;
TDAGNodeList = class;
TDAGEdgeList = class;

{==============================================================================}
{ TDAGNode declaration                                                         }
{==============================================================================}

TDAGNode = class( TGraphicControl )
private

	FEdges :                     TDAGEdgeList;    { Edges which connect this node to others }
    FBrushColor :                TColor;
    FBrushStyle :                TBrushStyle;
    FPenMode :                   TPenMode;
    FPenStyle :                  TPenStyle;
    FPenWidth :                  Integer;
    FSelected :                  Boolean;

    function  GetCenter :        TPoint ;

    procedure SetBrushColor(     NewBrushColor : TColor );
    procedure SetBrushStyle(     NewBrushStyle : TBrushStyle );

    procedure Select(            IsSelected :    Boolean );

    procedure InitPopup;
    procedure ChangeColor;
    procedure ZapNode;

    procedure   HandleMouseDown( Sender: TObject;
                                 Button: TMouseButton;
                                 Shift: TShiftState;
                                 X, Y: Integer );

    procedure   HandleMouseUp(   Sender: TObject;
                                 Button: TMouseButton;
                                 Shift: TShiftState;
                                 X, Y: Integer );

    procedure   HandleClick(     Sender : TObject );

    procedure   HandlePopupSel(  Sender : TObject );

    procedure   HandleDragOver(  Sender, Source: TObject;
                                 X, Y: Integer;
						    State: TDragState;
                                 var Accept: Boolean );

    procedure   HandleDragDrop(  Sender, Source: TObject;
                                 X, Y: Integer );

protected

    procedure   Paint; override;

public

	constructor Create(          AOwner : TComponent ); override;
    constructor Clone(           ASource : TDAGNode );
    destructor  Destroy; override;

    procedure BuildEdge(         toLink : TDAGNode );

	property    Center : TPoint read GetCenter;

published

	property Height default DAGNodeRadius*2;
	property Width default DAGNodeRadius*2;
    property Visible default True;
    property BrushColor : TColor read FBrushColor write SetBrushColor;
    property BrushStyle : TBrushStyle read FBrushStyle write SetBrushStyle;
    property Selected : Boolean read FSelected write Select;

end;

{==============================================================================}
{ TDAGEdge declaration                                                         }
{==============================================================================}

TDAGEdge = class( TGraphicControl )
private

	FAlpha :                     TDAGNode;
	FBeta :                      TDAGNode;
    FColor :                     TColor;
    FStyle :                     TPenStyle;
    FPenWidth :                  Integer;
    FDrawArrow :                 Boolean;

    procedure SetAlpha(          newAlpha : TDAGNode );
    procedure SetBeta(           newBeta : TDAGNode );
    procedure SetDimensions;
    procedure SetDrawArrow(      drawIt : Boolean );

    procedure HandleDragOver(    Sender, Source: TObject;
                                 X, Y: Integer;
						    State: TDragState;
                                 var Accept: Boolean );

    procedure HandleDragDrop(    Sender, Source: TObject;
                                 X, Y: Integer );
    procedure HandleClick(       Sender : TObject );

protected

    procedure Paint; override;

public

	constructor Create(          AOwner : TComponent ); override;
    destructor Destroy; override;

    procedure BindNodes(         newAlpha : TDAGNode;
                                 newBeta : TDAGNode );

published

	property Height default DAGNodeRadius*2;
	property Width default DAGNodeRadius*2;
    property Visible default True;

    property Color : TColor read FColor write FColor;
    property Style : TPenStyle read FStyle write FStyle;
    property PenWidth : Integer read FPenWidth write FPenWidth default DAGEdgeWidth;

    property Alpha : TDAGNode read FAlpha write SetAlpha;
    property Beta : TDAGNode read FBeta write SetBeta;
    property DrawArrow : Boolean read FDrawArrow write SetDrawArrow;

end;

{==============================================================================}
{ TDAGNodeList declaration                                                     }
{==============================================================================}

TDAGNodeList = class( TList )
private
	FOwnsItems :                 Boolean;	{ must destroy items in our Destroy? }
protected
public

	constructor Create;
    destructor Destroy; override;

	procedure AddNode(           node : TDAGNode );
    procedure RemoveNode(        node : TDAGNode ); { Remove w/o destroying }
    procedure ClearAndDestroy;

    property OwnsItems : Boolean read FOwnsItems write FOwnsItems default False;
end;

{==============================================================================}
{ TDAGEdgeList declaration                                                     }
{==============================================================================}

TDAGEdgeList = class( TList )
private
	FOwnsItems :                 Boolean;	{ must destroy items in our Destroy? }
protected
public

	constructor Create;
    destructor Destroy; override;

	procedure AddEdge(           edge : TDAGEdge );
    procedure RemoveEdge(        edge: TDAGEdge );  { Remove w/o destroying }
    procedure ClearAndDestroy;

    property OwnsItems : Boolean read FOwnsItems write FOwnsItems default False;
end;

{==============================================================================}
{ TDAGView declaration                                                         }
{==============================================================================}

TDAGView = class( TScrollBox )
private

	FNodes :                     TDAGNodeList;
    FEdges :                     TDAGEdgeList;
    FSelNode :                   TDAGNode;

    procedure HandleDragOver(    Sender, Source: TObject;
                                 X, Y: Integer;
						    State: TDragState;
                                 var Accept: Boolean );

    procedure HandleDragDrop(    Sender, Source: TObject;
                                 X, Y: Integer );

protected
public

	constructor Create(          AOwner : TComponent ); override;
    destructor Destroy; override;

	procedure AddNode(           node : TDAGNode );
    procedure RemoveNode(       node : TDAGNode );
    procedure AddEdge(           edge : TDAGEdge );
    procedure RemoveEdge(        edge : TDAGEdge );

    procedure ClearAll;
    procedure Redraw;

    procedure SelectNode(        node : TDAGNode );

    property SelNode : TDAGNode read FSelNode;

published
end;

procedure Register;

{##############################################################################}
	implementation
{##############################################################################}

var
	pupNodeMenu :            TPopupMenu;	{ Private popup used by nodes. }

{==============================================================================}
{ Debugging utilities                                                          }
{==============================================================================}

procedure Stub( stMethodName : String; stMessage: String );
begin
	MessageDlg( stMethodName + ' - ' + stMessage, mtInformation, [mbOK], 0 );
end;

procedure Beep( nTimes : integer );
var
	i, j :         LongInt;
begin
	for i := 1 to nTimes do begin

    	MessageBeep( Word( -1 ) );
         for j:= 1 to 1000000 do begin end;

    end;
end;

{==============================================================================}
{ TDAGNode implementation                                                      }
{==============================================================================}

constructor TDAGNode.Create( AOwner : TComponent );
begin
	inherited Create( AOwner );
    FEdges := TDAGEdgeList.Create;

    Height := DAGNodeRadius * 2;
    Width := DAGNodeRadius * 2;

    FBrushColor := clBlack;
    FBrushStyle := bsSolid;

    FPenMode := pmCopy;
    FPenStyle := psSolid;
    FPenWidth := 1;

    Visible := True;
    OnClick := HandleClick;
    OnMouseDown := HandleMouseDown;
    OnMouseUp := HandleMouseUp;
    OnDragOver := HandleDragOver;
    OnDragDrop := HandleDragDrop;

    if pupNodeMenu = Nil then
    	InitPopup;
end;

constructor TDAGNode.Clone( ASource : TDAGNode );
begin
	inherited Create( ASource.Owner );
    FEdges := TDAGEdgeList.Create;

    Height := ASource.Height;
    Width := ASource.Width;

    FBrushColor := ASource.FBrushColor;
    FBrushStyle := ASource.FBrushStyle;

    FPenMode := ASource.FPenMode;
    FPenStyle := ASource.FPenStyle;
    FPenWidth := ASource.FPenWidth;

    Visible := ASource.Visible;
    OnClick := HandleClick;
    OnMouseDown := HandleMouseDown;
    OnMouseUp := HandleMouseUp;
    OnDragOver := HandleDragOver;
    OnDragDrop := HandleDragDrop;
end;

destructor TDAGNode.Destroy;
var
	i :            integer;
    edge :         TDAGEdge;
begin
	FEdges.OwnsItems := True;
	FEdges.ClearAndDestroy;
    FEdges.Free;
    if Parent is TDAGView then (Parent as TDAGView).RemoveNode( self );
    inherited Destroy;
end;

function TDAGNode.GetCenter : TPoint;
begin
    GetCenter.X := Width div 2;
    GetCenter.Y := Height div 2;
end;

procedure TDAGNode.Select( IsSelected : Boolean );
begin
	FSelected := IsSelected;
    Refresh;
end;

procedure TDAGNode.SetBrushColor( NewBrushColor : TColor );
begin
	FBrushColor := NewBrushColor;
    Refresh;
end;

procedure TDAGNode.SetBrushStyle( NewBrushStyle : TBrushStyle );
begin
    FBrushStyle := NewBrushStyle;
    Refresh;
end;

procedure TDAGNode.Paint;
begin
	with Canvas do begin

         Pen.Mode := pmCopy;
    	Pen.Style := FPenStyle;

		if FSelected then begin
              Pen.Color := clWhite;
		end
		else begin
              Pen.Color := clBlack;
		end;

    	Brush.Color := FBrushColor;
    	Brush.Style := FBrushStyle;

		Ellipse( 0, 0, Width, Height );

    end;
end;

procedure TDAGNode.BuildEdge( toLink : TDAGNode );
var
	edge :                TDAGEdge;
    view :                TDAGView;
begin
	edge := TDAGEdge.Create( Owner );
    edge.Parent := Parent;

    if Parent is TDAGView then begin
    	view := Parent As TDAGView;
    	view.AddEdge( edge );
    end;

    edge.BindNodes( self, toLink );
end;

procedure TDAGNode.ChangeColor;
var
    dlgColor :           TColorDialog;
begin
	dlgColor := TColorDialog.Create( Owner );

    if dlgColor.Execute then
    	BrushColor := dlgColor.Color;

    dlgColor.free;
end;

procedure TDAGNode.ZapNode;
begin
	if Parent is TDAGView then
    	if (Parent as TDAGView).SelNode = self then
         	(Parent as TDAGView).SelectNode( Nil );
    Free;
end;

{ OnClick handler for private popup menu (internal, not user-settable.) }
procedure TDAGNode.HandlePopupSel( Sender : TObject );
var
	pit :                TMenuItem;
	node :               TDAGNode;
begin
	pit := Sender as TMenuItem;
   	node := (pit.Owner as TPopupMenu).PopupComponent as TDAGNode;

   	case( pit.Tag ) of
    	1 : node.ChangeColor;
         2 : node.ZapNode;
    else
    	MessageDlg( 'Unhandled popup event: ' + pit.Caption, mtInformation,
             			[mbOK], 0 );
    end;
end;

{ Setup private popup menu (should only be called once.) }
procedure TDAGNode.InitPopup;
var
	pit :                    TMenuItem;
begin

	if pupNodeMenu = Nil then begin

         pupNodeMenu := TPopupMenu.Create( Nil );

         pit := TMenuItem.Create( pupNodeMenu );
         pit.Caption := '&Color';
         pit.Tag := 1;
         pit.OnClick := HandlePopupSel;
         pupNodeMenu.Items.Add( pit );

         pit := TMenuItem.Create( pupNodeMenu );
         pit.Caption := '&Delete';
         pit.Tag := 2;
         pit.OnClick := HandlePopupSel;
    	pupNodeMenu.Items.Add( pit );

    end;
end;

{ OnMouseDown handler (internal, not user-settable.) }
procedure TDAGNode.HandleMouseDown(  Sender: TObject; Button: TMouseButton;
                                     Shift: TShiftState; X, Y: Integer );
var
    selected :               TDAGNode;
    pt :                     TPoint;
begin

    if ssLeft in Shift then

    	if (ssShift in Shift) then begin

         	if (Parent is TDAGView) then begin
				selected := (Parent As TDAGView).SelNode;
				if selected <> Nil then selected.BuildEdge( self );
			end

         end
         else BeginDrag( False )

    else if ssRight in Shift then

    	if pupNodeMenu <> Nil then begin

         	pt := ClientToScreen( Point( X, Y ) );

              pupNodeMenu.PopupComponent := self;
         	pupNodeMenu.Popup( pt.X, pt.Y );

         end;

end;


{ OnMouseUp handler (internal, not user-settable.) }
procedure TDAGNode.HandleMouseUp( Sender: TObject; Button: TMouseButton;
                                  Shift: TShiftState; X, Y: Integer );
begin
	EndDrag( Dragging );
end;

{ OnClick handler (internal, not user-settable.) }
procedure TDAGNode.HandleClick( Sender : TObject );
begin
    if not Dragging then
		if Parent is TDAGView then
    		(Parent as TDAGView).SelectNode( self );
end;

{ OnDragOver handler (internal, not user-settable. }
procedure TDAGNode.HandleDragOver( Sender, Source: TObject; X, Y: Integer;
							State: TDragState; var Accept: Boolean );
begin
	Accept := ( Source is TDAGNode );
end;

{ OnDragDrop handler (internal, not user-settable. }
procedure TDAGNode.HandleDragDrop( Sender, Source: TObject; X, Y: Integer );
var
	node :         TDAGNode;
begin

  	if Source is TDAGNode then begin

        	node := Source as TDAGNode;

		if node.Parent = self.Parent then
         	node.BuildEdge( self )
         else
         	MessageDlg( 'Cannot create edge with prototype -- '
              		+ 'drop it first on the viewer and then drag here!',
              			mtWarning, [mbOK], 0 );

    end;

end;


{==============================================================================}
{ TDAGEdge implementation                                                      }
{==============================================================================}

constructor TDAGEdge.Create( AOwner : TComponent );
begin
	inherited Create( AOwner );
    Height := DAGNodeRadius * 2;
    Width := DAGNodeRadius * 2;
    FDrawArrow := True;
    FPenWidth := DAGEdgeWidth;
    Visible := False;
    OnClick := HandleClick;
    OnDragOver := HandleDragOver;
    OnDragDrop := HandleDragDrop;
end;

destructor TDAGEdge.Destroy;
begin
	if Parent is TDAGView then (Parent as TDAGView).RemoveEdge( self );
	if FAlpha <> Nil then FAlpha.FEdges.RemoveEdge( self );	{ May break if already removed? }
    if FBeta <> Nil then FBeta.FEdges.RemoveEdge( self );
	inherited Destroy;
end;

procedure TDAGEdge.BindNodes( newAlpha : TDAGNode; newBeta : TDAGNode );
begin
	FAlpha := newAlpha;
    FBeta := newBeta;

    if FAlpha <> Nil then FAlpha.FEdges.AddEdge( self );
    if FBeta <> Nil then FBeta.FEdges.AddEdge( self );

    SetDimensions;
    Visible := True;
    SendToBack;
end;


procedure TDAGEdge.SetAlpha( newAlpha : TDAGNode );
begin
	if FAlpha <> Nil then
    	FAlpha.FEdges.RemoveEdge( self );

	FAlpha := newAlpha;

    if FAlpha <> Nil then
    	FAlpha.FEdges.AddEdge( self );

    SetDimensions;
end;

procedure TDAGEdge.SetBeta( newBeta : TDAGNode );
begin
	if FBeta <> Nil then
    	FBeta.FEdges.RemoveEdge( self );

	FBeta := newBeta;

    if FBeta <> Nil then
    	FBeta.FEdges.AddEdge( self );

    SetDimensions;
end;

procedure TDAGEdge.SetDimensions;
var
	pAlpha :          TPoint;
    pBeta :           TPoint;

	function min( n1, n2 : Integer ) : Integer;
	begin
		if n1 <= n2 then
    		min := n1
	    else
    		min := n2;
    end;

begin
	if (FAlpha <> Nil) and (FBeta <> Nil) then begin

	   	pAlpha.X := FAlpha.Left + FAlpha.Center.X;
	   	pAlpha.Y := FAlpha.Top + FAlpha.Center.Y;
    	pBeta.X := FBeta.Left + FBeta.Center.X;
    	pBeta.Y := FBeta.Top + FBeta.Center.Y;

	   	Left := min( pAlpha.X, pBeta.X );
    	Top := min( pAlpha.Y, pBeta.Y );
		Width := abs( pAlpha.X - pBeta.X );
    	Height := abs( pAlpha.Y - pBeta.Y );
    end;
end;

procedure TDAGEdge.SetDrawArrow( drawIt : Boolean );
begin
	FDrawArrow := drawIt;
    Refresh;
end;

procedure TDAGEdge.Paint;
var
	pAlpha, pBeta, pMid :            TPoint;

function midpoint( pFirst, pSecond : TPoint ) : TPoint;
begin
	midpoint.X := pFirst.X + trunc( (pSecond.X - pFirst.X) / 2 );
	midpoint.Y := pFirst.Y + trunc( (pSecond.Y - pFirst.Y) / 2 );
end;

procedure drawTriangle( Canvas : TCanvas; pCurrent, pDirection : TPoint; nSize : Integer );
var
	deltaX, deltaY :      Double;
	distance :            Double;
    dx, dy :              Integer;
	p1, p2, p3 :          TPoint;
begin
	deltaX := pDirection.X - pCurrent.X;
	deltaY := pDirection.Y - pCurrent.Y;
	distance := sqrt( (deltaX*deltaX) + (deltaY*deltaY) );

    dx := trunc( nSize * deltaX / distance );
    dy := trunc( nSize * deltaY / distance );

    p1.X := pCurrent.X - dy;
    p1.y := pCurrent.Y + dx;

    p2.X := pCurrent.X + (2 * dx);
    p2.Y := pCurrent.Y + (2 * dy);

    p3.X := pCurrent.X + dy;
    p3.Y := pCurrent.Y - dx;

    Canvas.Brush.Style := bsSolid;
    Canvas.Brush.Color := clBlack;
    Canvas.Polygon( [p1, p2, p3] );
end;

begin

	if FAlpha = Nil then begin
    	pAlpha.X := 0;
         pAlpha.Y := 0;
    end
    else	pAlpha := ScreenToClient( FAlpha.ClientToScreen( FAlpha.Center ) );

    if FBeta = Nil then begin
    	pBeta.X := Width;
         pBeta.Y := Height;
    end
    else pBeta := ScreenToClient( FBeta.ClientToScreen( FBeta.Center ) );

	with Canvas do begin

    	Brush.Style := bsClear;
         Pen.Color := FColor;
    	Pen.Style := FStyle;
         Pen.Width := FPenWidth;
         MoveTo( pAlpha.X, pAlpha.Y );
         LineTo( pBeta.X, pBeta.Y );
    end;

    if FDrawArrow then begin
    	pMid := midpoint( pAlpha, pBeta );
         DrawTriangle( Canvas, pMid, pBeta, FPenWidth * 4 );
    end;
end;

procedure TDAGEdge.HandleClick( Sender : TObject );
begin
end;

procedure TDAGEdge.HandleDragOver( Sender, Source: TObject; X, Y: Integer;
							State: TDragState; var Accept: Boolean );
begin
	Accept := (Source is TDAGNode);
end;

procedure TDAGEdge.HandleDragDrop( Sender, Source: TObject; X, Y: Integer );
var
    pt :              TPoint;
begin
	pt := parent.ScreenToClient( ClientToScreen( Point( X, Y ) ) );
    (parent as TDAGView).HandleDragDrop( Sender, Source, pt.X, pt.Y );
end;

{==============================================================================}
{ TDAGNodeList implementation                                                  }
{==============================================================================}

constructor TDAGNodeList.Create;
begin
	inherited Create;
    FOwnsItems := False;
end;

destructor TDAGNodeList.Destroy;
begin
	ClearAndDestroy;
    inherited Destroy;
end;

procedure TDAGNodeList.AddNode( node : TDAGNode );
begin
	Add( node );
end;

procedure TDAGNodeList.RemoveNode( node : TDAGNode );
begin
	Remove( node );
end;

procedure TDAGNodeList.ClearAndDestroy;
var
	i :          Integer;
    node :       TDAGNode;
begin

	if OwnsItems then

    	for i := Count - 1 downto 0 do begin
         	node := Items[i];	{ Grab the node,...               }
         	Delete( i );        { ...remove it from the list, ... }
              node.Free;		{ ...and destroy it.              }
         end;

    Clear;

end;

{==============================================================================}
{ TDAGEdgeList implementation                                                  }
{==============================================================================}


constructor TDAGEdgeList.Create;
begin
	inherited Create;
end;

destructor TDAGEdgeList.Destroy;
begin
	ClearAndDestroy;
    inherited Destroy;
end;


procedure TDAGEdgeList.AddEdge( edge : TDAGEdge );
begin
	Add( edge );
end;

procedure TDAGEdgeList.RemoveEdge( edge : TDAGEdge );
begin
	Remove( edge );
end;

procedure TDAGEdgeList.ClearAndDestroy;
var
	i :          Integer;
    edge :       TDAGEdge;
begin

	if OwnsItems then

    	for i := Count - 1 downto 0 do begin
         	edge := Items[i];	{ Grab the edge,...               }
         	Delete( i );        { ...remove it from the list, ... }
              edge.Free;		{ ...and destroy it.              }
         end;

    Clear;

end;

{==============================================================================}
{ TDAGView implementation                                                      }
{==============================================================================}

constructor TDAGView.Create( AOwner : TComponent );
begin
	inherited Create( AOwner );

    FNodes := TDAGNodeList.Create;
    FNodes.OwnsItems := True;

    FEdges := TDAGEdgeList.Create;
    FEdges.OwnsItems := True;

    DragCursor := crDrag;
    AutoScroll := True;
    OnDragOver := HandleDragOver;
    OnDragDrop := HandleDragDrop;
end;

destructor TDAGView.Destroy;
begin              
    FEdges.Destroy;
	FNodes.Destroy;
    inherited Destroy;
end;

procedure TDAGView.AddNode( node : TDAGNode );
begin
	FNodes.AddNode( node );
    node.Parent := self;
    node.Visible := True;
    node.Invalidate;
    Refresh;
end;

procedure TDAGView.RemoveNode( node : TDAGNode );
begin
	FNodes.RemoveNode( node );
end;

procedure TDAGView.AddEdge( edge : TDAGEdge );
begin
	FEdges.AddEdge( edge );
    edge.Parent := self;
    edge.Visible := True;
    edge.Invalidate;
    Refresh;
end;

procedure TDAGView.RemoveEdge( edge : TDAGEdge );
begin
	FEdges.RemoveEdge( edge );
end;

procedure TDAGView.Redraw;
var
	i :             Integer;
    edge :          TDAGEdge;
begin
	for i := 0 to FEdges.Count - 1 do begin
    	edge := FEdges[i];

         if edge <> Nil then begin
         	edge.SetDimensions;
			edge.Refresh;
         end;
    end;
end;

procedure TDAGView.ClearAll;
begin
	FEdges.ClearAndDestroy;
    FNodes.ClearAndDestroy;
    FSelNode := Nil;
    Refresh;
end;

procedure TDAGView.SelectNode( node : TDAGNode );
begin

   	if FSelNode <> Nil then
        	FSelNode.Select( False );

    if node <> Nil then
         node.Select( True );

    FSelNode := node;

end;

procedure TDAGView.HandleDragOver( Sender, Source: TObject; X, Y: Integer;
							State: TDragState; var Accept: Boolean );
begin
	Accept := ( Source is TDAGNode );
end;

procedure TDAGView.HandleDragDrop( Sender, Source: TObject; X, Y: Integer );
var
	node :         TDAGNode;
    edge :		TDAGEdge;
begin

  	if Source is TDAGNode then begin

        	node := Source as TDAGNode;

		if node.Parent <> self then
         	node := TDAGNode.Clone( node );

         node.Left := X;
         node.Top := Y;

		if node.Parent <> Sender then AddNode( node );

         SelectNode( node );

    end;

    Redraw;

end;

procedure Register;
begin
	RegisterComponents( 'DAG', [TDAGNode, TDAGEdge, TDAGView] );
end;

end.
